@@ -55,6 +55,7 @@ char *strncpy(char *dest, const char *src, size_t n) __THROW;
 #include	<asm/types.h>
 #include	<sys/ioctl.h>
 #define	MD_MAJOR 9
+#define MdpMinorShift 6
 
 #ifndef BLKGETSIZE64
 #define BLKGETSIZE64 _IOR(0x12,114,size_t) /* return device size in bytes (u64 *arg) */
@@ -73,12 +74,13 @@ enum mode {
 	MANAGE,
 	MISC,
 	MONITOR,
+	GROW,
 };
 
 extern char short_options[];
 extern struct option long_options[];
 extern char Version[], Usage[], Help[], OptionHelp[],
-	Help_create[], Help_build[], Help_assemble[],
+	Help_create[], Help_build[], Help_assemble[], Help_grow[],
 	Help_manage[], Help_misc[], Help_monitor[], Help_config[];
 
 /* structures read from config file */
@@ -93,20 +95,22 @@ extern char Version[], Usage[], Help[], OptionHelp[],
  */
 #define UnSet (0xfffe)
 typedef struct mddev_ident_s {
-	char *devname;
+	char	*devname;
 	
-	int uuid_set;
-	__u32 uuid[4];
+	int	uuid_set;
+	__u32	uuid[4];
 
 	unsigned int super_minor;
 
-	char *devices;		/* comma separated list of device
+	char	*devices;	/* comma separated list of device
 				 * names with wild cards
 				 */
-	int level;
+	int	level;
 	unsigned int raid_disks;
 	unsigned int spare_disks;
-	char *spare_group;
+	int	autof;		/* 1 for normal, 2 for partitioned */
+	char	*spare_group;
+
 	struct mddev_ident_s *next;
 } *mddev_ident_t;
 
@@ -132,11 +136,13 @@ struct mdstat_ent {
 	char		*level;
 	char		*pattern; /* U or up, _ for down */
 	int		percent; /* -1 if no resync */
+	int		resync; /* 1 if resync, 0 if recovery */
 	struct mdstat_ent *next;
 };
 
-extern struct mdstat_ent *mdstat_read(void);
+extern struct mdstat_ent *mdstat_read(int);
 extern void free_mdstat(struct mdstat_ent *ms);
+extern void mdstat_wait(int seconds);
 
 #ifndef Sendmail
 #define Sendmail "/usr/lib/sendmail -t"
@@ -144,15 +150,18 @@ extern void free_mdstat(struct mdstat_ent *ms);
 
 extern char *map_num(mapping_t *map, int num);
 extern int map_name(mapping_t *map, char *name);
-extern mapping_t r5layout[], pers[], modes[];
+extern mapping_t r5layout[], pers[], modes[], faultylayout[];
 
 extern char *map_dev(int major, int minor);
 
 
 extern int Manage_ro(char *devname, int fd, int readonly);
 extern int Manage_runstop(char *devname, int fd, int runstop);
+extern int Manage_resize(char *devname, int fd, long long size, int raid_disks);
+extern int Manage_reconfig(char *devname, int fd, int layout);
 extern int Manage_subdevs(char *devname, int fd,
 			  mddev_dev_t devlist);
+extern int Grow_Add_device(char *devname, int fd, char *newdev);
 
 
 extern int Assemble(char *mddev, int mdfd,
@@ -163,9 +172,9 @@ extern int Assemble(char *mddev, int mdfd,
 		    char *update,
 		    int verbose, int force);
 
-extern int Build(char *mddev, int mdfd, int chunk, int level,
+extern int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 		 int raiddisks,
-		 mddev_dev_t devlist);
+		 mddev_dev_t devlist, int assume_clean);
 
 
 extern int Create(char *mddev, int mdfd,
@@ -179,7 +188,7 @@ extern int Examine(mddev_dev_t devlist, int brief, int scan, int SparcAdjust);
 extern int Monitor(mddev_dev_t devlist,
 		   char *mailaddr, char *alert_cmd,
 		   int period, int daemonise, int scan, int oneshot,
-		   char *config, int test);
+		   char *config, int test, char *pidfile);
 
 extern int Kill(char *dev, int force);
 
@@ -190,6 +199,10 @@ extern int check_ext2(int fd, char *name);
 extern int check_reiser(int fd, char *name);
 extern int check_raid(int fd, char *name);
 
+extern int get_mdp_major(void);
+extern int is_standard(char *dev, int *nump);
+
+
 extern mddev_ident_t conf_get_ident(char *conffile, char *dev);
 extern mddev_dev_t conf_get_devs(char *conffile);
 extern char *conf_get_mailaddr(char *conffile);
@@ -215,3 +228,30 @@ extern void put_md_name(char *name);
 extern char *get_md_name(int dev);
 
 extern char DefaultConfFile[];
+
+extern int open_mddev(char *dev, int autof);
+
+
+#define	LEVEL_MULTIPATH		(-4)
+#define	LEVEL_LINEAR		(-1)
+#define	LEVEL_FAULTY		(-5)
+
+
+/* faulty stuff */
+
+#define	WriteTransient	0
+#define	ReadTransient	1
+#define	WritePersistent	2
+#define	ReadPersistent	3
+#define	WriteAll	4 /* doesn't go to device */
+#define	ReadFixable	5
+#define	Modes	6
+
+#define	ClearErrors	31
+#define	ClearFaults	30
+
+#define AllPersist	100 /* internal use only */
+#define	NoPersist	101
+
+#define	ModeMask	0x1f
+#define	ModeShift	5
