@@ -94,7 +94,7 @@ int md_get_version(int fd)
 	return  (vers.major*10000) + (vers.minor*100) + vers.patchlevel;
     if (errno == EACCES)
 	    return -1;
-    if (MAJOR(stb.st_rdev) == MD_MAJOR)
+    if (major(stb.st_rdev) == MD_MAJOR)
 	return (3600);
     return -1;
 }
@@ -150,6 +150,138 @@ int same_uuid(int a[4], int b[4])
     return 0;
 }
 
+void uuid_from_super(int uuid[4], mdp_super_t *super)
+{
+    uuid[0] = super->set_uuid0;
+    if (super->minor_version >= 90) {
+	uuid[1] = super->set_uuid1;
+	uuid[2] = super->set_uuid2;
+	uuid[3] = super->set_uuid3;
+    } else {
+	uuid[1] = 0;
+	uuid[2] = 0;
+	uuid[3] = 0;
+    }
+}
+
+int compare_super(mdp_super_t *first, mdp_super_t *second)
+{
+    /*
+     * return:
+     *  0 same, or first was empty, and second was copied
+     *  1 second had wrong number
+     *  2 wrong uuid
+     *  3 wrong other info
+     */
+    int uuid1[4], uuid2[4];
+    if (second->md_magic != MD_SB_MAGIC)
+	return 1;
+    if (first-> md_magic != MD_SB_MAGIC) {
+	memcpy(first, second, sizeof(*first));
+	return 0;
+    }
+
+    uuid_from_super(uuid1, first);
+    uuid_from_super(uuid2, second);
+    if (!same_uuid(uuid1, uuid2))
+	return 2;
+    if (first->major_version != second->major_version ||
+	first->minor_version != second->minor_version ||
+	first->patch_version != second->patch_version ||
+	first->gvalid_words  != second->gvalid_words  ||
+	first->ctime         != second->ctime         ||
+	first->level         != second->level         ||
+	first->size          != second->size          ||
+	first->raid_disks    != second->raid_disks    )
+	return 3;
+
+    return 0;
+}
+
+int load_super(int fd, mdp_super_t *super)
+{
+	/* try to read in the superblock
+	 *
+	 * return
+	 *   0 - success
+	 *   1 - no block size
+	 *   2 - too small
+	 *   3 - no seek
+	 *   4 - no read
+	 *   5 - no magic
+	 *   6 - wrong major version
+	 */
+	unsigned long size;
+	unsigned long long dsize;
+	unsigned long long offset;
+    
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = size << 9;
+	}
+
+	if (dsize < MD_RESERVED_SECTORS*2)
+		return 2;
+	
+	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
+
+	offset *= 512;
+
+	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */
+
+	if (lseek64(fd, offset, 0)< 0LL)
+		return 3;
+
+	if (read(fd, super, sizeof(*super)) != sizeof(*super))
+		return 4;
+
+	if (super->md_magic != MD_SB_MAGIC)
+		return 5;
+
+	if (super->major_version != 0)
+		return 6;
+	return 0;
+}
+
+int store_super(int fd, mdp_super_t *super)
+{
+	unsigned long size;
+	unsigned long long dsize;
+	
+	long long offset;
+    
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = ((unsigned long long)size) << 9;
+	}
+
+	if (dsize < MD_RESERVED_SECTORS*2)
+		return 2;
+	
+	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
+
+	offset *= 512;
+
+	if (lseek64(fd, offset, 0)< 0LL)
+		return 3;
+
+	if (write(fd, super, sizeof(*super)) != sizeof(*super))
+		return 4;
+
+	return 0;
+}
+    
+
 
 int check_ext2(int fd, char *name)
 {
@@ -208,24 +340,20 @@ int check_reiser(int fd, char *name)
 
 int check_raid(int fd, char *name)
 {
-	void *super;
-	struct mdinfo info;
+	mdp_super_t super;
 	time_t crtime;
-	struct supertype *st = guess_super(fd);
-
-	if (!st) return 0;
-	st->ss->load_super(st, fd, &super, name);
+	if (load_super(fd, &super))
+		return 0;
 	/* Looks like a raid array .. */
 	fprintf(stderr, Name ": %s appears to be part of a raid array:\n",
 		name);
-	st->ss->getinfo_super(&info, super);
-	free(super);
-	crtime = info.array.ctime;
+	crtime = super.ctime;
 	fprintf(stderr, "    level=%d devices=%d ctime=%s",
-		info.array.level, info.array.raid_disks, ctime(&crtime));
+		super.level, super.raid_disks, ctime(&crtime));
 	return 1;
 }
 
+
 int ask(char *mesg)
 {
 	char *add = "";
@@ -267,28 +395,37 @@ int map_name(mapping_t *map, char *name)
 }
 
 
-int is_standard(char *dev)
+int is_standard(char *dev, int *nump)
 {
 	/* tests if dev is a "standard" md dev name.
 	 * i.e if the last component is "/dNN" or "/mdNN",
 	 * where NN is a string of digits 
 	 */
-	dev = strrchr(dev, '/');
-	if (!dev)
+	char *d = strrchr(dev, '/');
+	int type=0;
+	int num;
+	if (!d)
 		return 0;
-	if (strncmp(dev, "/d",2)==0)
-		dev += 2;
-	else if (strncmp(dev, "/md", 3)==0)
-		dev += 3;
+	if (strncmp(d, "/d",2)==0)
+		d += 2, type=1; /* /dev/md/dN{pM} */
+	else if (strncmp(d, "/md_d", 5)==0)
+		d += 5, type=1; /* /dev/md_dNpM */
+	else if (strncmp(d, "/md", 3)==0)
+		d += 3, type=-1; /* /dev/mdN */
+	else if (d-dev > 3 && strncmp(d-2, "md/", 3)==0)
+		type = -1; /* /dev/md/N */
 	else
 		return 0;
-	if (!*dev)
+	if (!*d)
 		return 0;
-	while (isdigit(*dev))
-		dev++;
-	if (*dev)
+	num = atoi(d);
+	while (isdigit(*d))
+		d++;
+	if (*d)
 		return 0;
-	return 1;
+	if (nump) *nump = num;
+
+	return type;
 }
 
 
@@ -328,8 +465,8 @@ int add_dev(const char *name, const struct stat *stb, int flag)
 	char *n = strdup(name);
 	struct devmap *dm = malloc(sizeof(*dm));
 	if (dm) {
-	    dm->major = MAJOR(stb->st_rdev);
-	    dm->minor = MINOR(stb->st_rdev);
+	    dm->major = major(stb->st_rdev);
+	    dm->minor = minor(stb->st_rdev);
 	    dm->name = n;
 	    dm->next = devlist;
 	    devlist = dm;
@@ -340,7 +477,7 @@ int add_dev(const char *name, const struct stat *stb, int flag)
 
 /*
  * Find a block device with the right major/minor number.
- * Avoid /dev/mdNN and /dev/md/dNN is possible
+ * Avoid /dev/mdNN and /dev/md/dNN if possible
  */
 char *map_dev(int major, int minor)
 {
@@ -358,7 +495,7 @@ char *map_dev(int major, int minor)
 	for (p=devlist; p; p=p->next)
 		if (p->major == major &&
 		    p->minor == minor) {
-			if (is_standard(p->name))
+			if (is_standard(p->name, NULL))
 				std = p->name;
 			else
 				return p->name;
@@ -368,25 +505,19 @@ char *map_dev(int major, int minor)
 
 #endif
 
-unsigned long calc_csum(void *super, int bytes)
+unsigned long calc_sb_csum(mdp_super_t *super)
 {
+        unsigned int  oldcsum = super->sb_csum;
 	unsigned long long newcsum = 0;
+	unsigned long csum;
 	int i;
-	unsigned int csum;
-	unsigned int *superc = (unsigned int*) super;
+	unsigned int *superc = (int*) super;
+	super->sb_csum = 0;
 
-	for(i=0; i<bytes/4; i++)
+	for(i=0; i<MD_SB_BYTES/4; i++)
 		newcsum+= superc[i];
 	csum = (newcsum& 0xffffffff) + (newcsum>>32);
-#ifdef __alpha__
-/* The in-kernel checksum calculation is always 16bit on 
- * the alpha, though it is 32 bit on i386...
- * I wonder what it is elsewhere... (it uses and API in
- * a way that it shouldn't).
- */
-	csum = (csum & 0xffff) + (csum >> 16);
-	csum = (csum & 0xffff) + (csum >> 16);
-#endif
+	super->sb_csum = oldcsum;
 	return csum;
 }
 
@@ -478,14 +609,14 @@ char *get_md_name(int dev)
 	if (dev < 0) {
 		int mdp =  get_mdp_major();
 		if (mdp < 0) return NULL;
-		rdev = MKDEV(mdp, (-1-dev)<<6);
+		rdev = makedev(mdp, (-1-dev)<<6);
 		sprintf(devname, "/dev/md/d%d", -1-dev);
 		if (stat(devname, &stb) == 0
 		    && (S_IFMT&stb.st_mode) == S_IFBLK
 		    && (stb.st_rdev == rdev))
 			return devname;
 	} else {
-		rdev = MKDEV(MD_MAJOR, dev);
+		rdev = makedev(MD_MAJOR, dev);
 		sprintf(devname, "/dev/md%d", dev);
 		if (stat(devname, &stb) == 0
 		    && (S_IFMT&stb.st_mode) == S_IFBLK
@@ -498,7 +629,7 @@ char *get_md_name(int dev)
 		    && (stb.st_rdev == rdev))
 			return devname;
 	}
-	dn = map_dev(MAJOR(rdev), MINOR(rdev));
+	dn = map_dev(major(rdev), minor(rdev));
 	if (dn)
 		return dn;
 	sprintf(devname, "/dev/.tmp.md%d", dev);
@@ -519,64 +650,3 @@ void put_md_name(char *name)
 	if (strncmp(name, "/dev/.tmp.md", 12)==0)
 		unlink(name);
 }
-
-
-
-
-struct superswitch *superlist[] = { &super0, &super1, NULL };
-
-struct supertype *super_by_version(int vers, int minor)
-{
-	struct supertype *st = malloc(sizeof(*st));
-	if (!st) return st;
-	if (vers == 0)
-		st->ss = &super0;
-
-	if (vers == 1)
-		st->ss = &super1;
-	st->minor_version = minor;
-	return st;
-}
-
-struct supertype *guess_super(int fd)
-{
-	/* try each load_super to find the best match,
-	 * and return the best superswitch
-	 */
-	struct superswitch  *ss;
-	struct supertype *st;
-	unsigned long besttime = 0;
-	int bestsuper = -1;
-	
-	void *sbp = NULL;
-	int i;
-
-	st = malloc(sizeof(*st));
-	memset(st, 0, sizeof(*st));
-	for (i=0 ; superlist[i]; i++) {
-		int rv;
-		ss = superlist[i];
-		rv = ss->load_super(st, fd, &sbp, NULL);
-		if (rv == 0) {
-			struct mdinfo info;
-			ss->getinfo_super(&info, sbp);
-			if (bestsuper == -1 ||
-			    besttime < info.array.ctime) {
-				bestsuper = i;
-				besttime = info.array.ctime;
-				st->ss = NULL;
-			}
-			free(sbp);
-		}
-	}
-	if (bestsuper != -1) {
-		int rv;
-		rv = superlist[bestsuper]->load_super(st, fd, &sbp, NULL);
-		if (rv == 0) {
-			free(sbp);
-			return st;
-		}
-	}
-	free(st);
-	return NULL;
-}
